''
'' C3/Synapse memory driver - user interface
''
''        Author: Marko Lukat
'' Last modified: 2012/10/24
''       Version: 0.12
''
'' acknowledgements
'' - clkfreq/8 long assembler courtesy of Jonathan Dummer (lonesock)
''
'' 20121010: func_4[_unsafe], added 4n block read
'' 20121011: func_5[_unsafe], added 4n block write
'' 20121015: all functions verified on real h/w
'' 20121017: minor speed up in command interface (SPIN)
'' 20121018: corrected comments for block transfers (4n)
'' 20121020: func_6[_unsafe], added 8n block read
'' 20121022: added note re: startup sequence
'' 20121023: func_7[_unsafe], added 8n block write
'' 20121024: embed setn calls for simple block transfers (less overhead)
''
OBJ
  client: "core.con.client.C3"
  system: "core.con.system"

  helper: "helper"
  
VAR
  long  link[res_m]
  
PUB null
'' This is not a top level object.

PUB init(ID, mailbox, secondary{boolean})

  long[mailbox]{0} := TRUE
  long[mailbox][1] := result := system.launch(-1, @driver, mailbox)
'
' When using the driver in dual-cog mode, the secondary relies on
' its first read from long[mailbox]{0} being 0 (just after having
' been cleared by the primary). Assuming worst case (coginit same
' hub window, S after P) this is currently guaranteed by the length
' of their respective setup sequences:
'
' - P: 3 hub windows before clear
' - S: 5 hub windows before initial read
'
' Sequential startup would side-step this issue but wastes time.
'
  if secondary
    result.word[1] := system.launch(-1, helper.base, mailbox)
    repeat while long[mailbox][1]
  repeat while long[mailbox]{0}

PUB exec(parameters, command)

  command.word[1] := parameters
  link{0} := command
  repeat
  while link{0}

  return @link{0}
  
PUB SRAM_rdByte(src)

  exec(@result[1], cmd_rdByte)

PUB SRAM_wrByte(dst, value)

  exec(@result[1], cmd_wrByte)

PUB SRAM_rdBlock(src, length, dst{hub})

  exec(@result[1], cmd_rdBlock)

PUB SRAM_wrBlock(dst, length, src{hub})

  exec(@result[1], cmd_wrBlock)

CON
  cmd_rdByte    = $100A
  cmd_wrByte    = $3012
  cmd_rdBlock   = $5018
  cmd_wrBlock   = $5038

  cmd_rdBlock4n = $5058
  cmd_wrBlock4n = $5088

  cmd_rdBlock8n = $58B4
  cmd_wrBlock8n = $58E4
  
DAT {
'        function does(1) not(0) need secondary cog ------+
'                 function has(1) no(0) argument(s) ----+ |
'                            number of arguments -1 --+ | | 
'                                                     | | |
cmd_rdByte_     word    (@func_0 - @driver) >> 2 | %000_1_0 << 11 ' 1 argument
cmd_wrByte_     word    (@func_1 - @driver) >> 2 | %001_1_0 << 11 ' 2 arguments
cmd_rdBlock_    word    (@func_2 - @driver) >> 2 | %010_1_0 << 11 ' 3 arguments
cmd_wrBlock_    word    (@func_3 - @driver) >> 2 | %010_1_0 << 11 ' 3 arguments
cmd_rdBlock4n_  word    (@func_4 - @driver) >> 2 | %010_1_0 << 11 ' 3 arguments
cmd_wrBlock4n_  word    (@func_5 - @driver) >> 2 | %010_1_0 << 11 ' 3 arguments
cmd_rdBlock8n_  word    (@func_6 - @driver) >> 2 | %010_1_1 << 11 ' 3 arguments
cmd_wrBlock8n_  word    (@func_7 - @driver) >> 2 | %010_1_1 << 11 ' 3 arguments
}               
DAT             org     0                       ' cog binary header

header_2048     long    system#ID_2             ' magic number for a cog binary
                word    header_size             ' header size
                word    system#MAPPING          ' flags
                word    0, 0                    ' start register, register count

                word    @__table - @header_2048 ' translation table byte offset

header_size     fit     16
                
DAT             org     0                       ' memory driver

driver          jmpret  $, #setup               ' once

{done}          wrlong  zero, par
{idle}          rdlong  code, par wz
                test    code, argn wc           ' check for arguments
        if_z    jmp     #$-2

                mov     addr, code              ' args:n:s:[!Z]:cmd = 16:4:1:2:9
                ror     addr, #16               ' extract argument location
        if_c    call    #args                   ' fetch arguments
        if_c    addx    addr, #3                ' advance beyond last argument
                jmp     code                    ' execute function

' result.byte{0} := SRAM[arg0]

func_0          call    #setn           {%000}  ' set SRAM address
                andn    dira, #$FF
                movd    outa, #X10      {%010}  ' RD
                movs    arg1, ina               ' get data
                movd    outa, #X00      {%000}  ' IDLE

                sub     addr, #8                ' |
                wrbyte  arg1, addr              ' result.byte{0} := ina[7..0]
                jmp     %%0                     ' return

' SRAM[arg0] := arg1.byte{0}

func_1          call    #setn           {%000}  ' set SRAM address
                movs    outa, arg1              ' set data
                movd    outa, #X01      {%001}  ' WR
                movd    outa, #X00      {%000}  ' IDLE
                andn    dira, #$FF                
                jmp     %%0                     ' return

' repeat arg1
'   byte[arg2++] := SRAM[arg0++]

func_2          mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block

:loop   if_z    mov     arg6, #256              ' 1..256                        (%%)
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

{setn:s}        movi    outa, #%10_0_00 {%100}  ' reset-only in 23..31
                or      dira, #$FF

                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                movd    outa, #X10      {%110}  ' load  0..7  (data before command)
                shr     outa, #2        {%101}  ' load 16..23 (data after command)

                ror     arg0, #16
                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                andn    outa, down      {%100}  ' load 8..15 (data after command)

                movi    frqb, #%10000000_0      ' clkfreq/2

                shr     arg0, #24               ' SPI_SEL_CLK line free
                movs    outa, arg0
{setn:e}        mov     frqb, #0        {%000}

                add     arg7, #256              ' next block
                mov     arg0, arg7              ' transfer parameter

                andn    dira, #$FF

:transfer       cmp     arg6, #1 wz             '  +0 = n[RD-CNT-]RD-IDLE       (%%)

                movd    outa, #X10      {%010}  '  +4   RD
                movs    arg5, ina               '  +8   temp := SRAM[arg0]
        if_ne   movd    outa, #X11      {%011}  '  -4   CNT

                wrbyte  arg5, arg2              '  +0 = byte[arg2] := temp
                add     arg2, #1                '  +8   advance address
        if_ne   djnz    arg6, #:transfer        '  -4   for each byte

                movd    outa, #X00      {%000}  ' back to IDLE
                tjnz    arg1, #:loop            ' until done

                jmp     %%0                     ' return

' repeat arg1
'   SRAM[arg0++] := byte[arg2++]

func_3          mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block

:loop   if_z    mov     arg6, #256              ' 1..256                        (%%)
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

{setn:s}        movi    outa, #%10_0_00 {%100}  ' reset-only in 23..31
                or      dira, #$FF

                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                movd    outa, #X10      {%110}  ' load  0..7  (data before command)
                shr     outa, #2        {%101}  ' load 16..23 (data after command)

                ror     arg0, #16
                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                andn    outa, down      {%100}  ' load 8..15 (data after command)

                movi    frqb, #%10000000_0      ' clkfreq/2

                shr     arg0, #24               ' SPI_SEL_CLK line free
                movs    outa, arg0
{setn:e}        mov     frqb, #0        {%000}

                add     arg7, #256              ' next block
                mov     arg0, arg7              ' transfer parameter

:transfer       rdbyte  arg5, arg2              '  +0 = temp := byte[arg2]
                add     arg2, #1                '  +8   advance address

                cmp     arg6, #1 wz             '  -4   n[WR-CNT-]WR-IDLE       (%%)
                
                movs    outa, arg5              '  +0 = SRAM[arg0] := temp
                movd    outa, #X01      {%001}  '  +4   WR
        if_ne   movd    outa, #X11      {%011}  '  +8   CNT
        if_ne   djnz    arg6, #:transfer        '  -4   for each byte

                movd    outa, #X00      {%000}  ' back to IDLE
                tjnz    arg1, #:loop            ' until done

                andn    dira, #$FF                
                jmp     %%0                     ' return

' arg0 &= -4
' arg2 &= -4
'
' repeat arg1/4
'   repeat n from 0 to 3
'     temp.byte[n] := SRAM[arg0++]
'   long[arg2 += 4][-1] := temp

func_4          andn    arg1, #3 wz             ' |
        if_z    jmp     %%0                     ' nothing to transfer

                andn    arg0, #3                ' arg0 &= -4
'{wrlong}       andn    arg2, #3                ' arg2 &= -4
        
func_4_unsafe   mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block

                shr     arg1, #2                ' |
                shr     arg6, #2                ' long quantities

                movs    ctra, #AUTO_INC_RD      ' connect counter
                
:loop   if_z    mov     arg6, #64               ' 1..64                         (%%)
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

                call    #setn           {%000}  ' set SRAM address

                add     arg7, #256              ' next block
                mov     arg0, arg7              ' transfer parameter

                andn    dira, #$FF
                movd    outa, #X10      {%010}  ' switch to RD
                
:transfer       cmp     arg6, #1 wz             '  +0 = last long is treated    (%%)
        if_e    jmp     #:manually              '  +4   specially (early stop)

                movi    frqa, #%001_000000      '  +8   start
                movs    arg5, ina               '  -4
                ror     arg5, #8 wc             '  +0 =
                movs    arg5, ina               '  +4
                ror     arg5, #8                '  +8   clkfreq/8, "pulling a lonesock"
                movs    arg5, ina               '  -4
                rol     arg5, #15               '  +0 =
                movi    arg5, ina               '  +4
                movi    frqa, #%000_000000      '  +8   stop
                rcl     arg5, #1                '  -4
                
                wrlong  arg5, arg2              '  +0 = long[arg2] := temp
                add     arg2, #4                '  +8   advance address
                djnz    arg6, #:transfer        '  -4   for each long but the last

:manually       movi    frqa, #%001_000000      '  +8   start
                movs    arg5, ina               '  -4
                ror     arg5, #8 wc             '  +0 =
                movs    arg5, ina               '  +4
                ror     arg5, #8                '  +8   clkfreq/8, "pulling a lonesock"
                movs    arg5, ina               '  -4
                movi    frqa, #%000_000000      '  +0 = stop
                rol     arg5, #15               '  +4
                movi    arg5, ina               '  +8
                rcl     arg5, #1                '  -4

                wrlong  arg5, arg2              '  +0 = long[arg2] := temp
                add     arg2, #4                '  +8   advance address

                movd    outa, #X00      {%000}  ' back to IDLE
                tjnz    arg1, #:loop            ' until done

                jmp     %%0                     ' return

' arg0 &= -4
' arg2 &= -4
'
' repeat arg1/4
'   temp := long[arg2 += 4][-1]
'   repeat n from 0 to 3
'     SRAM[arg0++] := temp.byte[n]

func_5          andn    arg1, #3 wz             ' |
        if_z    jmp     %%0                     ' nothing to transfer

                andn    arg0, #3                ' arg0 &= -4
'{rdlong}       andn    arg2, #3                ' arg2 &= -4
        
func_5_unsafe   mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block

                shr     arg1, #2                ' |
                shr     arg6, #2                ' long quantities

                movs    ctra, #AUTO_INC_WR      ' connect counter
                
:loop   if_z    mov     arg6, #64               ' 1..64                         (%%)
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

                call    #setn           {%000}  ' set SRAM address

                add     arg7, #256              ' next block
                mov     arg0, arg7              ' transfer parameter

                movd    outa, #X01      {%001}  ' switch to WR

:transfer       rdlong  arg5, arg2              '  +0 = temp := long[arg2]
                add     arg2, #4                '  +8   advance address

                cmp     arg6, #1 wz             '  -4   last long is treated    (%%)
        if_e    jmp     #:manually              '  +0 = specially (early stop)
        
                movi    frqa, #%001_000000      '  +4   start
                movs    outa, arg5              '  +8
                shr     arg5, #8                '  -4
                movs    outa, arg5              '  +0 =
                shr     arg5, #8                '  +4
                movs    outa, arg5              '  +8
                shr     arg5, #8                '  -4
                movs    outa, arg5              '  +0 =
                movi    frqa, #%000_000000      '  +4   stop
                
                djnz    arg6, #:transfer        '  +8   for each long but the last

:manually       movi    frqa, #%001_000000      '  +4   start
                movs    outa, arg5              '  +8
                shr     arg5, #8                '  -4
                movs    outa, arg5              '  +0 =
                shr     arg5, #8                '  +4
                movs    outa, arg5              '  +8
                movi    frqa, #%000_000000      '  -4   stop
                shr     arg5, #8                '  +0 =
                movs    outa, arg5              '  +4

                movd    outa, #X00      {%000}  ' back to IDLE
                tjnz    arg1, #:loop            ' until done

                andn    dira, #$FF                
                jmp     %%0                     ' return

' arg0 &= -8
' arg2 &= -4
'
' repeat arg1/8                         repeat arg1/8
'   repeat n from 0 to 3                  repeat n from 0 to 3
'     temp.byte[n] := SRAM[arg0 +0][n]      temp.byte[n] := SRAM[arg0 +4][n]
'   arg0 += 8                             arg0 += 8
'   long[arg2 += 8][-2] := temp           long[arg2 +=8][-1] := temp

func_6          andn    arg1, #7 wz             ' |
        if_z    jmp     %%0                     ' nothing to transfer

                andn    arg0, #7                ' arg0 &= -8
'{wrlong}       andn    arg2, #3                ' arg2 &= -4
        
func_6_unsafe   mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block

                long    helper#cmd_rdBlock8n    '  -4   nop
                wrlong  $-1, orbiter            '  +0 = sync secondary

                shr     arg1, #3                '  +8   |
                shr     arg6, #3                '  -4   double long quantities

                movs    ctra, #AUTO_INC_RD      '  +0 = connect counter
                
:loop   if_z    mov     arg6, #32               '  +4   1..32                   (%%)
                max     arg6, arg1              '  +8   limit against what's left
                sub     arg1, arg6              '  -4   update what's left

{setn:s}        movi    outa, #%10_0_00 {%100}  '  +0 = reset-only in 23..31
                or      dira, #$FF              '  +4

                movs    outa, arg0              '  +8   SPI_SEL_CLK held high by ctrb
                movd    outa, #X10      {%110}  '  -4   load  0..7  (data before command)
                shr     outa, #2        {%101}  '  +0 = load 16..23 (data after command)

                ror     arg0, #16               '  +4
                movs    outa, arg0              '  +8   SPI_SEL_CLK held high by ctrb
                andn    outa, down      {%100}  '  -4   load 8..15 (data after command)

                movi    frqb, #%10000000_0      '  +0 = clkfreq/2

                shr     arg0, #24               '  +4   SPI_SEL_CLK line free
                movs    outa, arg0              '  +8
{setn:e}        mov     frqb, #0        {%000}  '  -4

                add     arg7, #256              '  +0 = advance block index
                andn    dira, #$FF              '  +4

                movd    outa, #X10      {%010}  '  +8   switch to RD
                movi    frqa, #%001_000000      '  -4   start

:transfer       movs    arg5, ina               '  +0 =
                ror     arg5, #8 wc             '  +4
                movs    arg5, ina               '  +8
                ror     arg5, #8                '  -4   clkfreq/8, "pulling a lonesock"
                movs    arg5, ina               '  +0 =
                rol     arg5, #15               '  +4
                movi    arg5, ina               '  +8
                rcl     arg5, #1                '  -4

                wrlong  arg5, arg2              '  +0 = long[arg2] := temp              movs    arg5, ina
                                                '  +4                                   ror     arg5, #8 wc
                add     arg2, #8                '  +8   advance address                 movs    arg5, ina  
                cmp     arg6, #1 wz             '  -4                           (%%)    ror     arg5, #8   
        if_e    mov     arg0, arg7              '  +0 = transfer parameter              movs    arg5, ina  
        if_e    movi    frqa, #%000_000000      '  +4   stop                            rol     arg5, #15  
        if_e    movd    outa, #X00      {%000}  '  +8   back to IDLE                    movi    arg5, ina  
        if_ne   djnz    arg6, #:transfer        '  -4   for each long                   rcl     arg5, #1   
                                                                                        
                tjnz    arg1, #:loop            '  +0 = until done

                jmp     %%0                     ' return

' arg0 &= -8
' arg2 &= -4
'
' repeat arg1/8                         repeat arg1/8
'   temp := long[arg2 += 8][-2]           temp := long[arg2 +=8][-1]
'   repeat n from 0 to 3                  repeat n from 0 to 3
'     SRAM[arg0 +0][n] := temp.byte[n]      SRAM[arg0 +4][n] := temp.byte[n]
'   arg0 += 8                             arg0 += 8

func_7          andn    arg1, #7 wz             ' |
        if_z    jmp     %%0                     ' nothing to transfer

                andn    arg0, #7                ' arg0 &= -8
'{rdlong}       andn    arg2, #3                ' arg2 &= -4
        
func_7_unsafe   mov     arg7, arg0              ' setn consumes arg0
                andn    arg7, #$FF              ' 256n

                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block
        if_z    mov     arg6, #256              ' 1..256

                long    helper#cmd_wrBlock8n    '  -4   nop
                wrlong  $-1, orbiter            '  +0 = sync secondary

                shr     arg1, #3                '  +8   |
                shr     arg6, #3                '  -4   double long quantities

                movs    ctra, #AUTO_INC_WR      '  +0 = connect counter
                
:loop           max     arg6, arg1              '  +4   limit against what's left
                sub     arg1, arg6              '  +8   update what's left

{setn:s}        movi    outa, #%10_0_00 {%100}  '  -4   reset-only in 23..31
                or      dira, #$FF              '  +0 =

                movs    outa, arg0              '  +4   SPI_SEL_CLK held high by ctrb
                movd    outa, #X10      {%110}  '  +8   load  0..7  (data before command)
                shr     outa, #2        {%101}  '  -4   load 16..23 (data after command)

                ror     arg0, #16               '  +0 =
                movs    outa, arg0              '  +4   SPI_SEL_CLK held high by ctrb
                andn    outa, down      {%100}  '  +8   load 8..15 (data after command)

                movi    frqb, #%10000000_0      '  -4   clkfreq/2

                shr     arg0, #24               '  +0 = SPI_SEL_CLK line free
                movs    outa, arg0              '  +4
{setn:e}        mov     frqb, #0        {%000}  '  +8

                movd    outa, #X01      {%001}  '  -4   switch to WR
                                                '                                       shr     outa, #8
                                                '                                       nop               
:transfer       rdlong  arg5, arg2              '  +0 = temp := long[arg2]              shr     outa, #8
                                                '  +4                                   cmp     arg6, #1 wz
                add     arg2, #8                '  +8   advance address                 shr     outa, #8
                movi    frqa, #%001_000000      '  -4   start                   if_nz   djnz    arg6, #wrTransfer

                movs    outa, arg5              '  +0 =                         
                shr     arg5, #8                '  +4                           
                movs    outa, arg5              '  +8                           
                shr     arg5, #8                '  -4                           
                movs    outa, arg5              '  +0 =                         
                shr     arg5, #8                '  +4                           
                movs    outa, arg5              '  +8                           
                cmp     arg6, #1 wz             '  -4                           

                movs    outa, #0                '  +0 = release bus                     mov     outa, arg5
        if_ne   djnz    arg6, #:transfer        '  +4   for each long                   add     arg2, #8

                add     arg7, #256              '  +8   advance block index             shr     outa, #8
                mov     arg0, arg7              '  -4   transfer parameter              nop
                mov     arg6, #32               '  +0 = 1..32                           shr     outa, #8
                movi    frqa, #%000_00000       '  +4   stop                            cmp     arg6, #1 wz
                nop                             '  +8                                   shr     outa, #8
                movd    outa, #X00      {%000}  '  -4   back to idle            if_nz   djnz    arg6, #wrTransfer
                tjnz    arg1, #:loop            '  +0 = until done

                andn    dira, #$FF                
                jmp     %%0                     ' return

'                                  v
setn            movi    outa, #%10_0_00 {%100}  ' reset-only in 23..31
                or      dira, #$FF

                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                movd    outa, #X10      {%110}  ' load  0..7  (data before command)
                shr     outa, #2        {%101}  ' load 16..23 (data after command)

                ror     arg0, #16
                movs    outa, arg0              ' SPI_SEL_CLK held high by ctrb
                andn    outa, down      {%100}  ' load  8..15 (data after command)

                movi    frqb, #%10000000_0      ' clkfreq/2

' Now we need to clock 6 times to force the final %100 to %000 transition.

                shr     arg0, #24               ' SPI_SEL_CLK line free
                movs    outa, arg0
                mov     frqb, #0        {%000}

setn_ret        ret

' support code

args            rdlong  arg0, addr              ' read 1st argument
                cmpsub  addr, delta wc          ' [increment address and] check exit
        if_nc   jmpret  zero, args_ret nr,wc    ' cond: early return
        
                rdlong  arg1, addr              ' read 2nd argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg2, addr              ' read 3rd argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg3, addr              ' read 4th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg4, addr              ' read 5th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg5, addr              ' read 6th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg6, addr              ' read 7th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg7, addr              ' read 8th argument
'               cmpsub  addr, delta wc
'       if_nc   jmpret  zero, args_ret nr,wc

args_ret        ret

' initialised data and/or presets

delta           long    %001_0 << 28 | $FFFC    '  %10  deal with movi setup
                                                ' -(-4) address increment
argn            long    |< 12                   ' function does have arguments
down            long    %111_1 << 8             ' X?? -> X00, SPI_SEL_CLK free

orbiter         long    4                       ' sub command link

' Stuff below is re-purposed for temporary storage.

setup           movi    ctra, #%0_00110_000     ' |
                movi    phsa, #%010_000000      ' auto-incrementer

                movi    ctrb, #%0_00100_000     ' |
                movs    ctrb, #SPI_SEL_CLK      ' |
                movi    phsb, #%10000000_0      ' SPI select counter

                mov     dira, mask      {%100}  ' drive outputs
                movi    outa, #%00_1_00         ' revoke reset
'                                  ^
                add     orbiter, par            ' resolve sub command link
                jmp     %%0                     ' return

mask            long    SPI_MASK|SPI_SEL_MASK   ' pin I/O setup

                fit

' uninitialised data and/or temporaries

                org     setup

addr            res     1
code            res     1

arg0            res     1
arg1            res     1
arg2            res     1
arg3            res     1
arg4            res     1
arg5            res     1
arg6            res     1
arg7            res     1

tail            fit

DAT                                             ' translation table

__table         word    (@__names - @__table)/2

                word    cmd_rdByte
                word    cmd_wrByte
                word    cmd_rdBlock
                word    cmd_wrBlock
                word    cmd_rdBlock4n
                word    cmd_wrBlock4n
                word    cmd_rdBlock8n
                word    cmd_wrBlock8n

                word    res_m
                
__names         byte    "rdByte", 0
                byte    "wrByte", 0
                byte    "rdBlock", 0
                byte    "wrBlock", 0
                byte    "rdBlock4n", 0
                byte    "wrBlock4n", 0
                byte    "rdBlock8n", 0
                byte    "wrBlock8n", 0

                byte    "res_m", 0

CON
  zero         = $1F0                           ' par (dst only)
  
  X00          = %000                           ' |
  X01          = %001                           ' |
  X11          = %101                           ' |
  X10          = %100                           ' CLK|MOSI movd

  res_m        = 2                              ' UI support

CON
  SPI_MOSI     = client#SPI_MOSI
  SPI_CLK      = client#SPI_CLK

  SPI_SEL_CLK  = client#SPI_SEL_CLK
  SPI_SEL_CLR  = client#SPI_SEL_CLR

  SPI_MASK     = |< SPI_CLK | |< SPI_MOSI
  SPI_SEL_MASK = |< SPI_SEL_CLK | |< SPI_SEL_CLR

  AUTO_INC_RD  = SPI_MOSI
  AUTO_INC_WR  = SPI_CLK
  
DAT