''
'' C3/Synapse memory driver - helper functions
''
''        Author: Marko Lukat
'' Last modified: 2012/10/24
''       Version: 0.4
''
'' acknowledgements
'' - clkfreq/8 long assembler courtesy of Jonathan Dummer (lonesock)
''
'' 20121020: initial version (8n block read)
'' 20121023: added 8n block write
'' 20121024: code clean-up, exit path analysis for block write
''
PUB null
'' This is not a top level object.

PUB base

  return @helper

CON
  cmd_rdBlock8n = $001B
  cmd_wrBlock8n = $002F
  
DAT             org     0

helper          jmpret  $, #setup               ' once

{idle}          rdlong  code, par wz
                test    code, argn wc           ' check for arguments
        if_z    jmp     #$-2

                mov     addr, code              ' args:n:s:[!Z]:cmd = 16:4:1:2:9        mov     addr, code
                ror     addr, #16               ' extract argument location             ror     addr, #16 
        if_c    call    #args                   ' fetch arguments               if_c    call    #args     
                test    code, help wc           ' check for support request     if_c    addx    addr, #3  
        if_c    jmp     #main                   '                                       jmp     code      

' Simple commands fall through. We wait here until the primary clears
' the command location before scanning for new commands. Otherwise we
' may get out of sync by repeatedly fetching parameters.

sync            rdlong  code, par wz
        if_nz   jmp     #$-1
                jmp     %%0                     ' return

' This code provides the second half for the 8n block transfers. SRAM
' control happens entirely in the primary cog.

main            andn    arg1, #7 wz             ' |                                     andn    arg1, #7 wz       
        if_z    jmp     #sync                   ' nothing to transfer           if_z    jmp     %%0               
                                                                                                                  
                andn    arg0, #7                ' arg0 &= -8                            andn    arg0, #7          
'{wr/rdlong}    andn    arg2, #3                ' arg2 &= -4                            andn    arg2, #3          

main_unsafe     rdlong  code, cmds wz           ' |                                     mov     arg7, arg0
                                                ' |                                     andn    arg7, #$FF
                mov     cnt, cnt                ' |                                     neg     arg6, arg0
        if_z    jmp     #$-2                    ' wait for sub command                  and     arg6, #$FF wz
                                                '                                       hubop   sync
                wrlong  zero, cmds              ' clear sub command

                add     cnt, #9{14} + (128 - 14 - 8) - 80
                sub     cnt, diff
                
                neg     arg6, arg0              ' 256 - arg0.byte{0}
                and     arg6, #$FF wz           ' limit within block                                              
                                                                                 
                shr     arg1, #3                ' |                                                               
                shr     arg6, #3                ' double long quantities                                          

                add     arg2, #4                ' intrinsic offset
                jmp     code

                
rdBlock8n       cmp     diff, #4 wc             '    1: wrlong before add
        if_nc   sub     arg2, #8                ' 2..7: add before wrlong

rdLoop          add     cnt, #64 + 16           ' 5 hub windows to cover setn & Co

        if_z    mov     arg6, #32               ' 1..32
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

rdTransfer      waitcnt cnt, #64                ' sync point

                movs    arg5, ina               ' |
                ror     arg5, #8 wc             ' |
                movs    arg5, ina               ' |
                ror     arg5, #8                ' clkfreq/8, "pulling a lonesock"
                movs    arg5, ina               ' |
                rol     arg5, #15               ' |
                movi    arg5, ina               ' |
                rcl     arg5, #1                ' |

rdPatch         wrlong  arg5, arg2              ' add     arg2, #8
                add     arg2, #8                ' wrlong  arg5, arg2
                
                djnz    arg6, #rdTransfer wz    ' for each long
                tjnz    arg1, #rdLoop           ' until done

                jmp     %%0                     ' return

' exit path timing analysis
'
'       |               |               |               |               |               |
'   S:  +---+---+---+---+---+---+---+---+---          H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---+---        H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---+---      H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---+---    H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---+---  H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---+---H-------+---####+---####+---    H
'   S:  +---+---+---+---+---+---+---+---  H-------+---+---####+---####+---H
'   P:  H-------+---+---+---+---+---+---+---####+---    H-------
'       |               |               |               |               |               |
'                                                       |
'                                       clear command --+

wrBlock8n       or      dira, #$FF

wrLoop          add     cnt, #64 + 16           ' 5 hub windows to cover setn & Co
                
        if_z    mov     arg6, #32               ' 1..32
                max     arg6, arg1              ' limit against what's left
                sub     arg1, arg6              ' update what's left

wrTransfer      mov     outa, #0                ' release bus

                rdlong  arg5, arg2              ' temp := long[arg2]

                waitcnt cnt, #64                ' sync point
                
                mov     outa, arg5
                add     arg2, #8                ' advance address
                shr     outa, #8
                cmp     arg1, #1 wc
                shr     outa, #8
                cmp     arg6, #1 wz
                shr     outa, #8
        if_nz   djnz    arg6, #wrTransfer       ' for each long

                mov     outa, #0                ' release bus

        if_nc   tjnz    arg1, #wrLoop           ' until done

                andn    dira, #$FF                
                jmp     %%0                     ' return

' exit path timing analysis
'
'       |               |               |               |               | 
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---              H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---            H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---          H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---        H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---      H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---    H
'   S:  +---+---+---+---+---+---+---+---+---+---+---+---  H
'   P:  +---+---+---+---+---+---+---+---+---####+---+---H-------
'       |               |               |               |               |
'                                                       |
'                                       clear command --+
                
' support code
        
args            rdlong  arg0, addr              ' read 1st argument
                cmpsub  addr, delta wc          ' [increment address and] check exit
        if_nc   jmpret  zero, args_ret nr,wc    ' cond: early return
        
                rdlong  arg1, addr              ' read 2nd argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg2, addr              ' read 3rd argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg3, addr              ' read 4th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg4, addr              ' read 5th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg5, addr              ' read 6th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg6, addr              ' read 7th argument
                cmpsub  addr, delta wc
        if_nc   jmpret  zero, args_ret nr,wc

                rdlong  arg7, addr              ' read 8th argument
'               cmpsub  addr, delta wc
'       if_nc   jmpret  zero, args_ret nr,wc

args_ret        ret

' initialised data and/or presets

delta           long    %001_0 << 28 | $FFFC    '  %10  deal with movi setup
                                                ' -(-4) address increment
argn            long    |< 12                   ' function does have arguments
help            long    |< 11                   ' function needs support

cmds            long    4                       ' sub command link

' Stuff below is re-purposed for temporary storage.

setup           cogid   diff                    ' who are we?

                add     cmds, par               ' resolve sub command link
                rdlong  temp, cmds              ' primary cog ID (+1)
                sub     temp, #1                ' adjust ID

                sub     diff, temp              ' slot delta
                and     diff, #%111             ' 1..7
                shl     diff, #1                ' cycles
                cmp     diff, #2 wz             ' wrlong first (default)

        if_ne   xor     rdPatch+1, rdPatch+0
        if_ne   xor     rdPatch+0, rdPatch+1
        if_ne   xor     rdPatch+1, rdPatch+0    ' add first
        
                wrlong  zero, cmds              ' confirm startup

                jmp     %%0                     ' return

                fit

' uninitialised data and/or temporaries

                org     setup

addr            res     1
code            res     1

arg0            res     1
arg1            res     1
arg2            res     1
arg3            res     1
arg4            res     1
arg5            res     1
arg6            res     1
arg7            res     1

tail            fit

CON
  zero = $1F0                                   ' par (dst only)
  temp = $1FA                                   ' frqa
  diff = $1FB                                   ' frqb
  
DAT